-- message.lua

--[[
	todo:
	get set included into message.lua
	combine message.draw & message.play
	possibly insert cfg into s thereby avoiding set
	find out difference between msg.over & msg.done??
		over controlled by keyboard (to turn off typewritter effect)?
		done by completion of the display of all chars in msg
	]]

local mess = {}

function mess.load()
	 -- shortcuts
    lg = love.graphics
    lk = love.keyboard
    lt = love.timer
	get_time = lt.getTime
	mess.advanceKey = message.advanceKey or 'z'
    mess.next = false 		-- last key press
	step_over = {['|c']=10, ['|,']=2, ['|:']=2, ['|!']=2}	-- # spaces control codes use
	mess.defaults	= {
		wincol		= {  0, 80,160,255},
		msgspd		= 20,
		textsh		= true,
		x			= 100,
		y			= 100,
		w			= 200,
		wait		= nil,
		skip		= true
		}end

local function unformat( s )		-- removes formatting codes from string
    s = s:gsub('|n','')
    s = s:gsub('|,','')
    s = s:gsub('|:','')
    s = s:gsub('|!','')
    s = s:gsub('|c%x%x%x%x%x%x%x%x','')
    return s
	end

	local bloc = {}						-- break text into blocks
	bloc.__index = bloc

function bloc:new(s,x,y,c)
    local chrBlk = {}					-- default character block
    setmetatable( chrBlk, bloc )
    chrBlk.c = c or color.white
    local _,r,g,b,a,col
    chrBlk.s			= unformat(s)	-- s  = string of characters (control code removed)
    chrBlk.nc			= #chrBlk.s		-- nc = number of characters in string
    chrBlk.x,chrBlk.y	= x, y			-- xy = starting location of string?
    if s:sub(1,2) == '|c' then
        _,_,r,g,b,a = s:find('(%x%x)(%x%x)(%x%x)(%x%x)',3)
        chrBlk.c = {tonumber(r,16),tonumber(g,16),tonumber(b,16),tonumber(a,16)}
		end
    return chrBlk
	end

function bloc:draw( x, y, nc )		-- draw nc chars starting at x,y
    local nc = nc or -1
	local c  = self.c or color.white
	local sh = .1
    if set.textsh then				-- add character shadow (if desired)
		lg.setColor( {sh*c[1],sh*c[2],sh*c[3], c[4]})
        lg.print( self.s:sub( 1, nc ), x + self.x - 1, y + self.y - 1)
		end
    lg.setColor( c )
	lg.print( self.s:sub( 1, nc ), x + self.x, y + self.y )
	end

--[[The following formatting options are implemented:
	|cRRGGBBAA changes the text color. each two digit number is in hexadecimal
		i.e. "|cff0000ff is text is in red"
	|! asks the played to press the "z" key (or whatever key.ent is) before moving on
	|n starts a new line
	|, pauses the text for the time required to display 10 characters
	|: pauses the text for the time required to display 50 characters
	|fnn changes char size for that window
	]]
function mess.prepare( s, set )
	local lineCount = 1
	mess.completed = false
	mess.fontSize = 12
	local loc = s:find("|f")-- optional fontsize change
	if loc then
		mess.fontSize = tonumber( s:sub( loc + 2, loc + 3 ) )
		local t = s:sub( 1, loc - 1 )
		s = t..s:sub( loc + 4, #s )
		end
	local font = love.graphics.newFont(mess.fontSize)
	love.graphics.setFont(font)
    fontw = function( line ) return font:getWidth( line ) end
	lg.setFont( font )
    msg = {}
    local x, y		= 0, 0		-- current character position in window
    local xc, yc	= x, y
	local i, j	= 1, 1
	local LineInWindow = 1	-- current block count (line in window)
	local dx,dy
    local cc = color.white
	setmetatable( set, {__index = mess.defaults } )
    if set.w then		-- if window width provided (set.w), wrap text by inserting '|n' to ss (substring)
        s = s .. ' '
        local ss = ''
        local si, sj
        local bi, bj = 1, 1
        local b = unformat( s )		-- 'b' is string with control codes removed (for line wrapping)
        j = 1
        for i = 1, #s do
			 if s:sub(i,i+1) == '|n' then
            ss = ss .. s:sub(bi,i+1)
            j = j-2
            si = i+1
            sj = j
            bi = si+1
            bj = sj+1
			end
			if s:sub(i,i) == ' ' then				-- find space between words for break
                if si and sj and (fontw(b:sub(bj,j-1)) > set.w) then	-- there must be at least one ' ' before this
					ss = ss .. s:sub(bi,si-1) .. '|n'			-- add line break
                    bi = si+1
                    bj = sj+1
					end
                si, sj = i, j
				end
            local next2chars = s:sub( i, i+1 )
			if step_over[next2chars] then j = j - step_over[next2chars] end
            j = j + 1
			end
		s = ss..s:sub(bi,#s)		-- string now has line break code (|n) in it
		end
    local ps = string.rep(' ',10)	-- insert 10 spaces (10/s.msgspd delay)
    local pl = string.rep(' ',50)	-- insert 50 spaces (50/s.msgspd delay)
    s = s:gsub('|,',ps..'|,')		-- insert blanks (for delays)
    s = s:gsub('|:',pl..'|:')
    msg.b = {}						-- output block
    msg.us = unformat(s)			-- text without control codes
    msg.tt = #msg.us/set.msgspd		-- time for message display
    msg.wait = set.wait				-- amount time delay after display (seconds)
    msg.skip = set.skip				-- whether typewritter effect can be avoided
    i,j = 1,1
    while true do
        line = ''
        local code = s:sub(i,i+1)		-- start new block if code found
        if code == '|!' then
            line = s:sub(j,i-1)
            dx,dy = fontw(unformat(line)),0
			elseif code == '|c' then
				line = s:sub(j,i-1)
				dx,dy = fontw(unformat(line)),0
			elseif code == '|n' then
				line = s:sub(j,i-1)
				dx,dy = -xc,mess.fontSize
			lineCount = lineCount+1
			elseif code == '|,' then
				line = s:sub(j,i-1)
				dx,dy = fontw(unformat(line))-fontw(ps),0	-- remove delay spaces(10)
			elseif code == '|:' then
				line = s:sub(j,i-1)
				dx,dy = fontw(unformat(line))-fontw(pl),0	-- remobe delay spaces(50)
			end
        if i == #s then			-- no codes in msg (overrides previous settings)
            line = s:sub(j,i)
            dx,dy = fontw(unformat(line)),0
			end
        if line ~= '' then			-- code found and not end of msg
            msg.b[LineInWindow] = bloc:new(line,xc,yc,cc)
            cc = msg.b[LineInWindow].c					-- color
            msg.b[LineInWindow].inp = (code == '|!')	-- == true if 'z' key pressed
            if (code == '|!') then
                msg.b[LineInWindow].t0		= 0
                msg.b[LineInWindow].tset	= false
                msg.b[LineInWindow].done	= false
				end
            xc, yc = xc + dx, yc + dy
            LineInWindow, j = LineInWindow + 1, i
			end
        if i == #s then break end
        i = i + 1
		end				-- ***** END WHILE *****
    msg.done		= false
    msg.over		= false
    msg.started		= false
    msg.t0			= get_time()		-- start time
	set.h = ( lineCount + 2 ) * mess.fontSize
	end

function mess.draw( )
	if not msg.over then
		mess.window( set.x, set.y, set.w+mess.fontSize, set.h )	-- creates window
		mess.play()					-- display text
		end
	mess.next = false
	end

function mess.play( )
    local cnc = 0
	local nc = math.floor( set.msgspd * (get_time() - msg.t0))
    if not msg.started then msg.started = true end
    if msg.done and not msg.over then	-- check if we are not done
        if set.wait == nil then
            if mess.next then msg.over = true end
			elseif get_time()-msg.t0 > set.wait then msg.over = true
			end
		end
    if msg.over then mess.completed = true return end

    if mess.next and msg.skip and 	-- enter key pressed & skip OK
		(not msg.b[msg.ci].inp or (msg.b[msg.ci].inp and
		(not msg.b[msg.ci].tset or msg.b[msg.ci].done))) then
        for i = 1,#msg.b do
            if msg.b[i].inp then msg.b[i].done = true end
			end
        msg.done	= true
        msg.t0		= get_time()
		end

    for i = 1, #msg.b do	-- sequentially display lines of chars
        if nc <= (msg.b[i].nc + cnc) and not msg.done then -- refresh current lines
            msg.b[i]:draw(set.x+10,set.y+mess.fontSize,nc-cnc)
            msg.ci = i
            return
			else
				msg.b[i]:draw(set.x+10,set.y+mess.fontSize) -- display previous (completed) blocs (lines of text)
				cnc = cnc+msg.b[i].nc
				if not msg.b[i].inp and i == #msg.b and not msg.done then
					msg.done = true
					msg.t0 = get_time()
					break
					end
				if msg.b[i].inp and not msg.b[i].done then -- keypress
					msg.ci = i
					if msg.b[i].tset then
						if mess.next then -- if input key pressed
							msg.b[i].done = true
							if i == #msg.b and not msg.done then
								msg.done = true
								msg.t0 = get_time()
								break
							end
							msg.t0 = msg.t0 + (get_time()-msg.b[i].t0)
						end
						break
					else
                    msg.b[i].t0 = get_time()
                    msg.b[i].tset = true
                    break
					end
				end
			end
		end
	end


function mess.window(x,y,w,h)
    lg.setLine(1,'rough')
    lg.setColor(set.wincol)--color.window)
    lg.rectangle('fill',x+1,y+4,w,h-4)
    lg.setColor({255,255,255,64})
	lg.setLine(5,'rough')
    lg.rectangle('line',x+1,y+3,w,h-3)
	lg.setLine(1,'rough')
	end

return mess
